package lufyt.gui;

import java.text.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/**
 * Abstract extension of JPanel for a graphical object to set an
 * option. The value of the option can be extracted by the methods
 * getString and getNumber.
 */
public abstract class Option extends JPanel {
    
    /** Label for this option */
    protected JLabel label;

    /** Returns an enum identifier for the type of data set by this option. */
    public abstract Type getType();
    /** Returns a string representation of the value of this option. */
    public abstract String getString();
    /** Reset this option to its default value. */
    public abstract void reset();
    /** Clear this option. */
    public abstract void clear();
    /** Returns a number representation of the value of this option. */
    public Number getNumber() {return null;}
    /** Returns the label text and and the value for this option. */
    public String toString() {return label.getText() + " " + getString();}
    /** Returns an errormessage if the value of this option is invalid
     * and null if no error is found. */
    public String errorCheck() {return null;}

    /** Enum type for specification of the type of data set by this option */
    public static enum Type {STRING, INTEGER, REAL, BOOLEAN, DISCRETE}

    /** Recommended maximal width for this JPanel */
    public static final int maxWidth = 600;
    
    protected Option() {}
    /** Forwards layout to the constructor of JPanel. */
    protected Option(LayoutManager layout) {super(layout);}

    /** Concretization of Option that provides a JTextField for input. */
    public static class Field extends Option {
	protected JTextField field;
	protected Type type;
	protected String defaultString;
	public static final int maxWidth = 600;

	public Field(String label, String defString) {
	    this(label, defString, Type.STRING);
	}

	public Type getType() {return type;}

	/** Returns the text of the JTextField object of this option. */
	public String getString() {return field.getText();}

	/** Writes the default string to the JTextField of this option. */
	public void reset() {field.setText(defaultString);}

	/** Writes the empty string to the JTextField of this option. */
	public void clear() {field.setText("");}

	/** Constructor that specify the label string, the default
	 * string, and the option type. */
	protected Field(String label, String defString, Type t) {
	    super(new GridLayout(0, 2, 12, 0));
	    type = t;
	    defaultString = defString;
	    this.label = new JLabel(label, JLabel.RIGHT);
	    field = new JTextField(defString);
	    if(t != Type.STRING)
		field.setHorizontalAlignment(JTextField.RIGHT);
	    add(this.label);
	    setMaximumSize(new Dimension(maxWidth/2,
					 field.getPreferredSize().height));
	    add(field);
	}
    }

    /** Extention of Option.Field for numeric options. */
    public static class NumberField extends Field {
	protected NumberFormat numberFormat;
	protected Number min;
	protected Number max;
	protected Number defValue;
	protected boolean strictMin;
	protected boolean strictMax;

	/** Constructs an unbounded integer option field. */
	public NumberField(String label, Integer defValue) {
	    this(label, defValue, null, null);}

	public NumberField(String label, Integer defValue,
			   Integer min) {
	    this(label, defValue, min, null);}
	/** Constructs an integer option field with the provided
	 * inclusive bounds. The bounds min and max may be null
	 * objects to specify that no bound should be imposed. */
	public NumberField(String label, Integer defValue,
			   Integer min, Integer max) {
	    super(label, format(null, defValue), Type.INTEGER);
	    this.defValue = defValue;
	    this.min = min;
	    this.max = max;
	}

	/** Constructs an unbounded double option field. */
	public NumberField(String label, Double defValue) {
	    this(label, defValue, null, false, null, false);}

	/** Constructs a double option field with a lower bound. */
	public NumberField(String label, Double defValue,
			   Double min, boolean strictMin) {
	    this(label, defValue, min, strictMin, null, false);}

	/** Constructs an integer option field with the provided
	 * bounds. The bounds min and max may be null objects to
	 * specify that no bound should be imposed. */
	public NumberField(String label, Double defValue,
			   Double min, boolean strictMin,
			   Double max, boolean strictMax) {
	    super(label, format(null, defValue), Type.REAL);
	    this.defValue = defValue;
	    this.min = min;
	    this.max = max;
	    this.strictMin = strictMin;
	    this.strictMax = strictMax;
	}

	/** Sets the numeric format of this option. A null object
	 * yields the default behaviour. */
	public void setNumberFormat(NumberFormat nf) {
	    numberFormat = nf;
	    defaultString = format(nf, defValue.doubleValue());
	}

	/** Internal number-formatting method. */
	protected static String format(NumberFormat nf, double val) {
	    if(nf != null)
		return nf.format(val);
	    String str = String.format("%G", val);
	    char sep = (new DecimalFormatSymbols()).getDecimalSeparator();
	    char tSep = String.format("%G", 0.1).charAt(1);
	    if(sep != tSep) {
		int i = str.indexOf(tSep);
		if(i >= 0)
		    str = str.substring(0, i) + sep + str.substring(i + 1);
	    }
	    if(str.indexOf(sep) >= 0) {
		int e = str.indexOf('E');
		if(e < 0)
		    e = str.length();
		int i = e;
		while(i > 1 && str.charAt(i - 1) == '0')
		    --i;
		if(str.charAt(i - 1) == sep)
		    --i;
		if(i < e)
		    str = str.substring(0, i) + str.substring(e);
	    }
	    return str;
	}

	public Number getNumber() {
	    ParsePosition p = new ParsePosition(0);
	    String s = getString().toUpperCase();
	    Number n = (numberFormat != null? numberFormat:
			NumberFormat.getInstance()).parse(s, p);
	    return p.getIndex() == s.length()? n: null;
	}
	
	/**
	 * Check if the string of this option can be parsed into a
	 * number in the allowed range.
	 */
	public String errorCheck() {
	    Number n = getNumber();
	    if(n == null)
		return "Ogiltligt numeriskt format för " + label.getText();
	    String err = null;
	    if(type == Type.INTEGER) {
		if(min != null && n.intValue() < min.intValue())
		    err = "mindre än minsta tillåtna värde "
			+ format(numberFormat, min.intValue());
		else if(max != null && n.intValue() > max.intValue())
		    err = "större än största tillåtna värde "
			+ format(numberFormat, max.intValue());
	    }
	    else {
		if(min != null) {
		    if(n.doubleValue() < min.doubleValue())
			err = "mindre än minsta tillåtna värde "
			    + format(numberFormat, min.doubleValue());
		    else if(strictMin && n.doubleValue() == min.doubleValue())
			err = "lika med strikt nedre gräns "
			    + format(numberFormat, min.doubleValue());
		}
		else if(max != null) {
		    if(n.doubleValue() > max.doubleValue())
			err = "större än största tillåtna värde "
			    + format(numberFormat, max.intValue());
		    if(strictMax && n.doubleValue() == max.doubleValue())
			err = "lika med strikt övre gräns "
			    + format(numberFormat, max.doubleValue());
		}
	    }
	    if(err != null)
		return "Angivet tal är " + err + " för " + label.getText();
	    return null;
	}
    }

    /**
     * Concretization of Option that provides radio buttons for
     * selection between a few vallues.
     */
    public static class Discrete extends Option {
	protected JRadioButton[] buttonArr;
	protected String[] idArr;
	protected String[] itemArr;
	protected int defaultIndex;

	/** Convenience constructor that sets the default index to 0. */
	public Discrete(String label, int gap, String[] id, String[] item) {
	    this(label, gap, id, item, 0);}

	/**
	 * Constructs an integer option field with the provided id
	 * strings and item label strings. The parameter gap specifies
	 * additional padding between the option label and the radio
	 * button group.
	 */
	public Discrete(String label, int gap, String[] id, String[] item,
			int defaultIndex) {
	    setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	    this.label = new JLabel(label, JLabel.CENTER);
	    this.label.setAlignmentX(Component.CENTER_ALIGNMENT);
	    add(this.label);
	    if(gap > 0)
		add(Box.createRigidArea(new Dimension(0, gap)));
	    JPanel bPanel = new JPanel();
	    bPanel.setLayout(new BoxLayout(bPanel, BoxLayout.Y_AXIS));
	    bPanel.setAlignmentX(Component.CENTER_ALIGNMENT);
	    add(bPanel);
	    ButtonGroup group = new ButtonGroup();
	    buttonArr = new JRadioButton[id.length];
	    for(int i = 0; i < id.length; ++i) {
		buttonArr[i] = new JRadioButton(item[i]);
		group.add(buttonArr[i]);
		bPanel.add(buttonArr[i]);
	    }
	    idArr = (String[]) id.clone();
	    itemArr = (String[]) item.clone();
	    this.defaultIndex = defaultIndex;
	    reset();
	}

	/** Returns Type.DISCRETE. */
	public Type getType() {return Type.DISCRETE;}

	/** Returns the index of the selected item for this option. */
	public Number getNumber() {
	    for(int i = 0; i < buttonArr.length; ++i)
		if(buttonArr[i].isSelected())
		    return new Integer(i);
	    return null;
	}

	/** Returns the id string of the selected item for this option. */
	public String getString() {return idArr[getNumber().intValue()];}

	/** Selects the default item of this option. */
	public void reset() {buttonArr[defaultIndex].setSelected(true);}

	/** Selects the first item of this option. */
	public void clear() {buttonArr[0].setSelected(true);}
    }

    /** Untested extention of Option.Field for Boolean options. */
    public static class Toggle extends Option {
	protected JCheckBox button;
	protected boolean defaultValue;

	public Toggle(String label, boolean defVal) {
	    button = new JCheckBox(label, defVal);
	    defaultValue = defVal;
	}

	public Type getType() {return Type.BOOLEAN;}

	public Number getNumber() {
	    return new Integer(button.isSelected()? 1: 0);}

	public String getString() {
	    return button.isSelected()? "true": "false";}

	public void reset() {button.setSelected(defaultValue);}

	public void clear() {button.setSelected(false);}
    }

    /** JPanel that contains options as graphical components. */
    public static class Panel extends JPanel implements ActionListener {
	
	public Insets defaultInsets;

	protected Map<String, Option> optionMap;
	protected Vector<Option> optionVec;
	protected Map<String, JComponent> actionMap;
	protected Set<String> actionValidate;
	protected Set<ActionListener> actionListeners;
	protected String pressedId;
	protected String errorTitle;
	protected int buttonWidth;

	public Panel() {
	    setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	    optionMap = new HashMap<String, Option>();
	    optionVec = new Vector<Option>();
	    actionMap = new HashMap<String, JComponent>();
	    actionValidate = new HashSet<String>();
	    actionListeners = new HashSet<ActionListener>();
	    errorTitle = "Parameterfel";
	    defaultInsets = new Insets(3, 12, 3, 12);
	}

	/** Add an option with default insets to this panel. */
	public Option putOption(String id, Option opt) {
	    return putOption(id, opt, defaultInsets);}

	/** Add an option with specified insets to this panel. */
	public Option putOption(String id, Option opt, Insets ins) {
	    Option prev = optionMap.put(id, opt);
	    if(prev != null) {
		remove(prev);
		optionVec.remove(prev);
	    }
	    add(opt);
	    optionVec.add(opt);
	    opt.setBorder(new EmptyBorder(ins));
	    opt.setAlignmentX(Component.CENTER_ALIGNMENT);
	    return prev;
	}

	/** Add an activity button with automatic validation and
	 * default insets to this panel. */
	public JComponent putButton(String id, JButton button) {
	    return putButton(id, button, defaultInsets);}

	/** Add an activity button with default insets to this panel. */
	public JComponent putButton(String id, JButton button,
				    boolean validateOptions) {
	    return putButton(id, button, validateOptions, defaultInsets);}

	/** Add an activity button with automatic validation and
	 * specified insets to this panel. */
	public JComponent putButton(String id, JButton button, Insets ins) {
	    return putButton(id, button, true, ins);}

	/** Add an activity button with specified insets to this panel. */
	public JComponent putButton(String id, JButton button,
				    boolean validateOptions, Insets ins) {
	    JComponent c = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
	    JComponent prev = actionMap.put(id, c);
	    if(prev != null) {
		((JButton)prev.getComponent(0)).removeActionListener(this);
		remove(prev);
	    }
	    c.add(button);
	    add(c);
	    c.setMaximumSize
		(new Dimension(maxWidth, c.getPreferredSize().height));
	    c.setBorder(new EmptyBorder(ins));
	    if(validateOptions)
		actionValidate.add(id);
	    else
		actionValidate.remove(id);
	    button.addActionListener(this);
	    button.setActionCommand(id);
	    int w = button.getPreferredSize().width;
	    if(w > buttonWidth) {
		buttonWidth = w;
		for(JComponent p: actionMap.values()) {
		    JComponent b = (JComponent) p.getComponent(0);
		    b.setPreferredSize
			(new Dimension(buttonWidth,
				       b.getPreferredSize().height));
		}
	    }
	    else
		button.setPreferredSize
		    (new Dimension(buttonWidth,
				   button.getPreferredSize().height));
	    return prev;
	}

	/** Add a vertical gap in the layout of this panel. */
	public void addGap(int gap) {
	    if(gap > 0)
		add(Box.createRigidArea(new Dimension(0, gap)));
	}

	/** Add vertical glue in the layout of this panel. */
	public void addStretch() {add(Box.createVerticalGlue());};

	/** Add a label with default insets to the layout of this panel. */
	public void addLabel(String str) {addLabel(str, defaultInsets);}

	/** Add a label with specified insets to the layout of this panel. */
	public void addLabel(String str, Insets ins) {
	    addLabel(new JLabel(str, JLabel.CENTER), ins);}

	/** Add a label with default insets to the layout of this panel. */
	public void addLabel(JLabel label) {addLabel(label, defaultInsets);}

	/** Add a label with specified insets to the layout of this panel. */
	public void addLabel(JLabel label, Insets ins) {
	    add(label);
	    label.setBorder(new EmptyBorder(ins));
	    label.setAlignmentX(Component.CENTER_ALIGNMENT);
	}

	/** Calls getString() for the option with the given id. */
	public String getString(String id) {
	    Option opt = optionMap.get(id);
	    if(opt == null)
		return null;
	    return opt.getString();
	}

	/** Calls getNumber() for the option with the given id. */
	public Number getNumber(String id) {
	    Option opt = optionMap.get(id);
	    if(opt == null)
		return null;
	    return opt.getNumber();
	}

	/** Returns the id of the last pressed button. */
	public String lastPressed() {return pressedId;}

	public void addActionListener(ActionListener aL) {
	    actionListeners.add(aL);
	}
	
	public void removeActionListener(ActionListener aL) {
	    actionListeners.remove(aL);
	}

	public void errorDialog(String msg) {
	    JOptionPane.showMessageDialog(getRootPane(), msg, errorTitle,
					  JOptionPane.ERROR_MESSAGE);
	}

	public void setErrorTitle(String title) {errorTitle = title;}

	public String getErrorTitle() {return errorTitle;}

	/** Calls getNumber() for each option. */
	public String errorCheck() {
	    for(Option opt: optionVec) {
		String s = opt.errorCheck();
		if(s != null)
		    return s;
	    }
	    return null;
	}

	/** Extract a string representing each option. */
	public String[] getOptionText() {
	    String[] text = new String[optionVec.size()];
	    for(int i = 0; i < optionVec.size(); ++i)
		text[i] = optionVec.get(i).toString();
	    return text;
	}

	/** Press the button with the given id. */
   	public void pressButton(String id) {
	    JComponent cmp = actionMap.get(id);
	    if(cmp == null)
		return;
	    ((JButton)cmp.getComponent(0)).doClick();
	}

	public void actionPerformed(ActionEvent e) {
	    String id = e.getActionCommand();
	    if(actionValidate.contains(id))
		for(Option opt: optionVec) {
		    String err = opt.errorCheck();
		    if(err != null) {
			errorDialog(err + ".");
			return;
		    }
		}
	    pressedId = id;
	    for(ActionListener aL: actionListeners)
		aL.actionPerformed(e);
	}
    }
}
