package lufyt.plot;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;

/**
 * Abstract base-class for xy-data to be plotted by a Graphis2D object.
 */
public abstract class Graph {
    
    private Vector<Double> x = new Vector<Double>();
    private Vector<Double> y = new Vector<Double>();
    private Double xMin;
    private Double xMax;
    private Double yMin;
    private Double yMax;

    /**
     * Sets the color used for plotting. A null argument is valid and
     * tells this Graph to not change the current color of the
     * provided Graphic2D object. */
    public abstract void setColor(Color c);

    /**
     * Draws this graph on g2 using the transforms provided by the
     * Range.Reader arguments and the AffineTransform t. */
    public abstract void draw(Graphics2D g2, AffineTransform t,
			      Range.Reader x, Range.Reader y);

    /** Returns the minimal x-value of the data stored by this Graph. */
    public Double getXMin() {return xMin;}
    /** Returns the maximal x-value of the data stored by this Graph. */
    public Double getXMax() {return xMax;}
    /** Returns the minimal y-value of the data stored by this Graph. */
    public Double getYMin() {return yMin;}
    /** Returns the maximal y-value of the data stored by this Graph. */
    public Double getYMax() {return yMax;}

    /** Returns the total number of data-points stored by this Graph. */
    public int nPoints() {return x.size();}

    /** Returns a reader object for read-only access to this Graph. */
    public Reader reader() {return new Reader(this);}

    /** Removes all data points stored by this Graph */
    public void clear() {
	x.clear();
	y.clear();
	xMin = xMax = yMin = yMax = null;
    }

    /**
     * Appends a single data point to this Graph and returns true on
     * success. */
    public boolean add(double x, double y) {
	if(Double.isInfinite(x) || Double.isNaN(x) ||
	   Double.isInfinite(y) || Double.isNaN(y))
	    return false;
	if(this.x.size() == 0) {
	    xMin = xMax = x;
	    yMin = yMax = y;
	}
	else {
	    if(x < xMin) xMin = x;
	    if(x > xMax) xMax = x;
	    if(y < yMin) yMin = y;
	    if(y > yMax) yMax = y;
	}
	this.x.add(x);
	this.y.add(y);
	return true;
    }

    /**
     * Appends data points to this Graph from the provided Collections
     * until the end of at least one Collection is reached. Returns
     * the number of spprnfrf data points. */
    public int add(Collection<Double> xData, Collection<Double> yData) {
	if(xData == null || yData == null)
	    return 0;
	Iterator<Double> xIter;
	Iterator<Double> yIter;
	int n = 0;
	for(xIter = xData.iterator(), yIter = yData.iterator();
	    xIter.hasNext() && yIter.hasNext(); )
	    if(add(xIter.next(), yIter.next()))
		++n;
	return n;
    }

    /**
     * Class to provide read-only access for a current Graph
     * object. This class is final to prevent acess to data-altering
     * methods of the current Graph. */
    public static final class Reader {
	private Graph graph;
	public Reader() {}
	/** Constructor that sets g to the current Graph of this object. */
	public Reader(Graph g) {graph = g;}
	/** Sets g to the current Graph of this object. */
	public void set(Graph g) {graph = g;}
	/** Delegates a draw function call to the current Graph. */
	public void draw(Graphics2D g2, AffineTransform t,
			 Range.Reader x, Range.Reader y) {
	    graph.draw(g2, t, x, y);}
	/** Delegates a getXMin function call to the current Graph. */
	public Double getXMin() {return graph.getXMin();}
	/** Delegates a getXMax function call to the current Graph. */
	public Double getXMax() {return graph.getXMax();}
	/** Delegates a getXMin function call to the current Graph. */
	public Double getYMin() {return graph.getYMin();}
	/** Delegates a getXMax function call to the current Graph. */
	public Double getYMax() {return graph.getYMax();}
	/** Delegates a getXMin function call to the current Graph. */
	public int nPoints() {return graph.nPoints();}
    }
 
    /**
     * JComponent for drawing of a collection of graphs by override
     * of the paint method. */
    public static class Panel extends JComponent {
    	/** A collection of Graph readers to be plotted on this object. */
	public Collection<Graph.Reader> content = new Vector<Graph.Reader>();
	private Range.Reader xRange;
	private Range.Reader yRange;

	/** Constuctor that sets the Range readers to be used by this object. */
	public Panel(Range.Reader x, Range.Reader y) {
	    super();
	    xRange = x;
	    yRange = y;
	}

	public boolean hasGraphs() {
	    return content != null && content.iterator().hasNext();
	}

	/** Returns the minimal x-value of the data referenced by content. */
	public Double getXMin() {
	    if(!hasGraphs()) return null;
	    Iterator<Graph.Reader> it = content.iterator();
	    double xMin = it.next().getXMin();
	    while(it.hasNext())
		xMin = Math.min(xMin, it.next().getXMin());
	    return xMin;
	}

	/** Returns the maximal x-value of the data referenced by content. */
	public Double getXMax() {
	    if(!hasGraphs()) return null;
	    Iterator<Graph.Reader> it = content.iterator();
	    double xMax = it.next().getXMax();
	    while(it.hasNext())
		xMax = Math.max(xMax, it.next().getXMax());
	    return xMax;
	}

	/** Returns the minimal y-value of the data referenced by content. */
	public Double getYMin() {
	    if(!hasGraphs()) return null;
	    Iterator<Graph.Reader> it = content.iterator();
	    double yMin = it.next().getYMin();
	    while(it.hasNext())
		yMin = Math.min(yMin, it.next().getYMin());
	    return yMin;
	}

	/** Returns the maximal y-value of the data referenced by content. */
	public Double getYMax() {
	    if(!hasGraphs()) return null;
	    Iterator<Graph.Reader> it = content.iterator();
	    double yMax = it.next().getYMax();
	    while(it.hasNext())
		yMax = Math.max(yMax, it.next().getYMax());
	    return yMax;
	}

	/**
	 * Calculates the affine transform that is to be used in
	 * combination with the transforms of the Range readers
	 * provided by the constructor. */
	public AffineTransform getAffineTransform() {
	    final double x0 = xRange.transform(xRange.getMin());
	    final double x1 = xRange.transform(xRange.getMax());
	    final double y0 = yRange.transform(yRange.getMin());
	    final double y1 = yRange.transform(yRange.getMax());
	    AffineTransform t = new AffineTransform();
	    t.translate(0.5, getHeight() - 0.5);
	    t.scale((getWidth() - 1)/(x1 - x0),
		    (1 - getHeight())/(y1 - y0));
	    t.translate(-x0, -y0);
	    return t;
	}

	/** Paints the graphs refenced by content. */
	public void paint(Graphics g) {
	    Graphics2D g2 = (Graphics2D) g;
	    if(hasGraphs()) {
		g2.setRenderingHints(new RenderingHints
				     (RenderingHints.KEY_ANTIALIASING,
				      RenderingHints.VALUE_ANTIALIAS_ON));
		for(Iterator<Graph.Reader> i = content.iterator(); i.hasNext();)
		    i.next().draw(g2, getAffineTransform(), xRange, yRange);
	    }
	    super.paint(g);
	}
    }

    /** Concrete extension of Graph that draws a line-plot. */
    public static class Line extends Graph {
	private Color color;
	private static double pixelCutoff = 3e4;

	/** Set the color of this plot. */
	public void setColor(Color c) {color = c;}

	public void draw(Graphics2D g2, AffineTransform t,
			 Range.Reader xT, Range.Reader yT) {
	    Iterator<Double> xIter = super.x.iterator();
	    Iterator<Double> yIter = super.y.iterator();
	    if(!xIter.hasNext())
		return;
	    if(color != null)
		g2.setColor(color);
	    GeneralPath path = new GeneralPath();
	    Point2D src = new Point2D.Double();
	    Point2D dest = new Point2D.Double();
	    do {
		src.setLocation(xT.transform(xIter.next()),
				yT.transform(yIter.next()));
		t.transform(src, dest);
		if(Math.abs(dest.getX()) < pixelCutoff &&
		   Math.abs(dest.getY()) < pixelCutoff)
		    path.moveTo(dest.getX(), dest.getY());
	    } while(xIter.hasNext() && (Math.abs(dest.getX()) > pixelCutoff ||
					Math.abs(dest.getY()) > pixelCutoff));
	    while(xIter.hasNext()) {
		src.setLocation(xT.transform(xIter.next()),
				yT.transform(yIter.next()));
		t.transform(src, dest);
		if(Math.abs(dest.getX()) < pixelCutoff &&
		   Math.abs(dest.getY()) < pixelCutoff)
		    path.lineTo(dest.getX(), dest.getY());
	    }
	    g2.draw(path);
	}
    }

    /** Concrete extension of Graph that draws a ring for each data point. */
    public static class Rings extends Graph {
	private Color color;
	private GeneralPath ring;
	private boolean filled;
	private static double pixelCutoff = 3e4;

	/**
	 * Constructor that sets the ring radius to r and specifies
	 * whether the rings should be filled.
	 */
	public Rings(double r, boolean fill) {
	    ring = new GeneralPath(new Ellipse2D.Double(-r, -r, 2*r, 2*r));
	    filled = fill;
	}

	/** Set the color of this plot. */
	public void setColor(Color c) {color = c;}

	public void draw(Graphics2D g2, AffineTransform t,
			 Range.Reader xT, Range.Reader yT) {
	    Iterator<Double> xIter = super.x.iterator();
	    Iterator<Double> yIter = super.y.iterator();
	    if(!xIter.hasNext())
		return;
	    if(color != null)
		g2.setColor(color);
	    GeneralPath path = new GeneralPath(ring);
	    Point2D src = new Point2D.Double();
	    Point2D dest = new Point2D.Double();
	    AffineTransform offs = new AffineTransform();
	    while(xIter.hasNext()) {
		src.setLocation(xT.transform(xIter.next()),
				yT.transform(yIter.next()));
		t.transform(src, dest);
		if(Math.abs(dest.getX()) < pixelCutoff
		   && Math.abs(dest.getY()) < pixelCutoff) {
		    offs.setToTranslation(dest.getX(), dest.getY());
		    if(filled)
			g2.fill(ring.createTransformedShape(offs));
		    else
			g2.draw(ring.createTransformedShape(offs));
		}
	    }
	}
    }
}
