package lufyt.plot;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;

/** Plotpanel for displaying graphs in a plot-area with labeled axes. */
public class Plot extends JLayeredPane {

    public Graph.Panel graphPanel;
    public ScalePanel scalePanel;

    protected JComponent title;
    protected final Axis xAxis = new Axis();
    protected final Axis yAxis = new Axis();
    protected int labelMargin;
    protected int titleMargin;
    protected int paneMargin;

    public void setBoundsX(Double min, Double max) {xAxis.setBounds(min, max);}
    public void setBoundsY(Double min, Double max) {yAxis.setBounds(min, max);}
    public void setScaleX(Range r) {xAxis.set(r);}
    public void setScaleY(Range r) {yAxis.set(r);}
    public void setScaleX(Range r, Scale s) {xAxis.set(r, s);}
    public void setScaleY(Range r, Scale s) {yAxis.set(r, s);}
    public void set(Range rX, Range rY, Double minX, Double minY,
		    Double maxX, Double maxY) {
	setScaleX(rX);
	setScaleY(rY);
	setBoundsX(minX, maxX);
	setBoundsY(minY, maxY);
    }
    public void setLinX() {setScaleX(new Range.Lin());}
    public void setLinY() {setScaleY(new Range.Lin());}
    public void setLogX() {setScaleX(new Range.Log());}
    public void setLogY() {setScaleY(new Range.Log());}
    public void setLinLin(Double x0, Double y0, Double x1, Double y1) {
	set(new Range.Lin(), new Range.Lin(), x0, y0, x1, y1);}
    public void setLinLog(Double x0, Double y0, Double x1, Double y1) {
	set(new Range.Lin(), new Range.Log(), x0, y0, x1, y1);}
    public void setLogLin(Double x0, Double y0, Double x1, Double y1) {
	set(new Range.Log(), new Range.Lin(), x0, y0, x1, y1);}
    public void setLogLog(Double x0, Double y0, Double x1, Double y1) {
	set(new Range.Log(), new Range.Log(), x0, y0, x1, y1);}

    public void setTitle(JComponent title) {
        replace(this.title, title, 0);
	this.title = title;
    }
    public void setLabelX(JComponent label) {
	replace(xAxis.label, label, 0);
	xAxis.label = label;
    }
    public void setLabelY(JComponent label) {
	replace(yAxis.label, label, 0);
	yAxis.label = label;
    }
    /**
     * Specifies the horizontal gap between the x- and y-labels and
     * the border of the plot-area.*/
    public void setLabelMargin(int m) {labelMargin = m;}
    public void setTitleMargin(int m) {titleMargin = m;}
    public void setPaneMargin(int m)  {paneMargin = m;}

    public Plot() {
	graphPanel = new Graph.Panel(xAxis.rangeReader, yAxis.rangeReader);
	graphPanel.setBorder(BorderFactory.createLineBorder(Color.black));
	scalePanel = new ScalePanel(graphPanel, xAxis, yAxis);
	add(scalePanel, new Integer(0));
	add(graphPanel, new Integer(1));
	setLinX();
	setLinY();
	setPaneMargin(16);
	setPreferredSize(new Dimension(600, 480));
	setMinimumSize(new Dimension(80, 80));
    }

    public void doLayout() {
	if(graphPanel.content != null && graphPanel.content.size() != 0) {
	    xAxis.range.min = xAxis.min == null? graphPanel.getXMin(): xAxis.min;
	    xAxis.range.max = xAxis.max == null? graphPanel.getXMax(): xAxis.max;
	    yAxis.range.min = yAxis.min == null? graphPanel.getYMin(): yAxis.min;
	    yAxis.range.max = yAxis.max == null? graphPanel.getYMax(): yAxis.max;
	    xAxis.scale.setTics();
	    yAxis.scale.setTics();
	    
	    int w = getWidth();
	    int h = getHeight();
	    Dimension titleDim = getDimension(title);
	    Dimension xLabelDim = getDimension(xAxis.label);
	    Dimension yLabelDim = getDimension(yAxis.label);
	    
	    int top = paneMargin + Math.max(yLabelDim.height + labelMargin,
					    titleDim.height + titleMargin);
	    int bott = paneMargin + xLabelDim.height + labelMargin;
	    if(title != null)
		title.setBounds((w - titleDim.width)/2,
				top - titleMargin - titleDim.height,
				titleDim.width, titleDim.height);
	    if(xAxis.label != null)
		xAxis.label.setBounds(w - paneMargin - xLabelDim.width,
				      h - bott + labelMargin,
				      xLabelDim.width, xLabelDim.height);
	    if(yAxis.label != null)
	    yAxis.label.setBounds(paneMargin,
				  top - labelMargin - yLabelDim.height,
				  yLabelDim.width, yLabelDim.height);
	    scalePanel.setBounds(paneMargin, top,
				 w - 2*paneMargin, h - top - bott);
	}
	super.doLayout();
    }

    /** Class for for handling of label, scale, and range of a plot-axis. */
    public static class Axis {
	public JComponent label;
	public final Range.Reader rangeReader = new Range.Reader(null);
	protected Scale scale;
	protected Range range;
	protected Double min;
	protected Double max;

	/** Non-null arguments fix the bounds of this axis. */
	void setBounds(Double min, Double max) {
	    this.min = min;
	    this.max = max;
	}
	/** Set the range and and use default scale. */
	void set(Range r) {set(r, r.getDefaultScale());}
	/** Set the range and the scale that this axis reads. */
	void set(Range r, Scale s) {
	    rangeReader.set(range = r);
	    (scale = s).setReader(rangeReader);
	}
    }

    /** JComponent that draws plot-axes with labels, scales, and tics. */
    public static class ScalePanel extends JComponent {
	protected Graph.Panel graph;
	protected Axis xAxis;
	protected Axis yAxis;
	protected int leftSpace;
	protected int rightSpace;
	protected int bottomSpace;
	protected int topSpace;
	protected int minorTic;
	protected int majorTic;
	protected boolean cachedLayout;
	protected int xAxisTextHeight;
	protected int yAxisTextWidth;
	protected int width;
	protected int height;

	public ScalePanel(Graph.Panel graph, Axis xAxis, Axis yAxis) {
	    this.graph = graph;
	    this.xAxis = xAxis;
	    this.yAxis = yAxis;
	    setSpaces(9, 15, 6, 6);
	    setTics(5, 3);
	}

	/** Sets all margins around this plot-area including labels and title. */
	public void setSpaces(int s) {setSpaces(s, s, s, s);}

	/** Sets the margins around this plot-area including labels and title. */
	public void setSpaces(int left, int right, int bottom, int top) {
	    if(leftSpace != left || rightSpace != right ||
	       bottomSpace != bottom || topSpace != top) {
		leftSpace = left;
		rightSpace = right;
		bottomSpace = bottom;
		topSpace = top;
		cachedLayout = false;
	    }
	}

	/** Sets the lenghs of major and minor scale-tics. */
	public void setTics(int major, int minor) {
	    majorTic = major;
	    minorTic = minor;
	}

	/** Validates internal layout and paints this object. */
	public void paint(Graphics g) {
	    if(graph.content == null || graph.content.size() == 0) {
		super.paint(g);
		return;
	    }

	    Vector<Double> xMajor = new Vector<Double>();
	    Vector<Double> xMinor = new Vector<Double>();
	    Vector<Double> yMajor = new Vector<Double>();
	    Vector<Double> yMinor = new Vector<Double>();
	    xAxis.scale.getTics(xMajor, xMinor);
	    yAxis.scale.getTics(yMajor, yMinor);
	    String[] xText = new String[xMajor.size()];
	    String[] yText = new String[yMajor.size()];

	    //Validate internal layout
	    FontMetrics fm = getFontMetrics(getFont());
	    int tH = fm.getAscent();
	    for(int i = 0; i < xText.length; ++i)
		xText[i] = Scale.numberString(xMajor.get(i));
	    int yW = 0;
	    for(int i = 0; i < yText.length; ++i) {
		yText[i] = Scale.numberString(yMajor.get(i));
		yW = Math.max(yW, fm.stringWidth(yText[i]));
	    }
	    //Update layout if invalid
	    if(!cachedLayout || width != getWidth() || height != getHeight()
	       || tH != xAxisTextHeight || yW != yAxisTextWidth) {
		width = getWidth();
		height = getHeight();
		xAxisTextHeight = tH;
		yAxisTextWidth = yW;
		graph.setBounds(getX() + leftSpace + yW, getY() + topSpace,
				width - yW - leftSpace - rightSpace,
				height - topSpace - bottomSpace - tH);
		cachedLayout = true;
	    }

	    //Setup scale-tics
	    AffineTransform t = new AffineTransform();
	    t.translate(graph.getX() - getX(), graph.getY() - getY());
	    t.concatenate(graph.getAffineTransform());
	    double preX0 = xAxis.range.transform(xAxis.range.min);
	    double preY0 = yAxis.range.transform(yAxis.range.min);
	    double preX1 = xAxis.range.transform(xAxis.range.max);
	    double preY1 = yAxis.range.transform(yAxis.range.max);
	    Point2D pre = new Point2D.Double();
	    Point2D post0 = new Point2D.Double();
	    pre.setLocation(preX0, preY0);
	    t.transform(pre, post0);
	    Point2D post1 = new Point2D.Double();
	    pre.setLocation(preX1, preY1);
	    t.transform(pre, post1);
	    GeneralPath xTic = new GeneralPath();
	    xTic.moveTo(0, post0.getY());
	    xTic.lineTo(0, post0.getY() - majorTic);
	    xTic.moveTo(0, post1.getY() + majorTic);
	    xTic.lineTo(0, post1.getY());
	    GeneralPath yTic = new GeneralPath();
	    yTic.moveTo(post0.getX(), 0);
	    yTic.lineTo(post0.getX() + majorTic, 0);
	    yTic.moveTo(post1.getX() - majorTic, 0);
	    yTic.lineTo(post1.getX(), 0);

	    //Setup graphics object
	    Graphics2D g2 = (Graphics2D) g;
	    g2.setRenderingHints(new RenderingHints
				 (RenderingHints.KEY_ANTIALIASING,
				  RenderingHints.VALUE_ANTIALIAS_ON));

	    //Draw tics and scales
	    Point2D post = new Point2D.Double();
	    AffineTransform d = new AffineTransform();
	    for(int i = 0; i < xText.length; ++i) {
		pre.setLocation(xAxis.range.transform(xMajor.get(i)), preY0);
		t.transform(pre, post);
		g2.drawString(xText[i],
			      (int)(post.getX() - fm.stringWidth(xText[i])/2.),
			      (int)(post.getY() + bottomSpace + tH));
		d.setToIdentity();
		d.translate(post.getX(), 0);
		GeneralPath p = (GeneralPath)xTic.clone();
		p.transform(d);
		g2.draw(p);
	    }
	    for(int i = 0; i < yText.length; ++i) {
		pre.setLocation(preX0, yAxis.range.transform(yMajor.get(i)));
		t.transform(pre, post);
		g2.drawString(yText[i],
			      (int)(post.getX() - fm.stringWidth(yText[i])
				    - leftSpace),
			      (int)(post.getY() + tH/2.));
		d.setToIdentity();
		d.translate(0, post.getY());
		GeneralPath p = (GeneralPath)yTic.clone();
		p.transform(d);
		g2.draw(p);
	    }
	    xTic.reset();
	    xTic.moveTo(0, post0.getY());
	    xTic.lineTo(0, post0.getY() - minorTic);
	    xTic.moveTo(0, post1.getY() + minorTic);
	    xTic.lineTo(0, post1.getY());
	    yTic.reset();
	    yTic.moveTo(post0.getX(), 0);
	    yTic.lineTo(post0.getX() + minorTic, 0);
	    yTic.moveTo(post1.getX() - minorTic, 0);
	    yTic.lineTo(post1.getX(), 0);
	    for(int i = 0; i < xMinor.size(); ++i) {
		pre.setLocation(xAxis.range.transform(xMinor.get(i)), preY0);
		t.transform(pre, post);
		d.setToIdentity();
		d.translate(post.getX(), 0);
		GeneralPath p = (GeneralPath)xTic.clone();
		p.transform(d);
		g2.draw(p);
	    }
	    for(int i = 0; i < yMinor.size(); ++i) {
		pre.setLocation(preX0, yAxis.range.transform(yMinor.get(i)));
		t.transform(pre, post);
		d.setToIdentity();
		d.translate(0, post.getY());
		GeneralPath p = (GeneralPath)yTic.clone();
		p.transform(d);
		g2.draw(p);
	    }

	    super.paint(g);
	}
    }
    
    private void replace(JComponent cOld, JComponent cNew, int layer) {
	if(cOld != null) remove(cOld);
	if(cNew != null) add(cNew, new Integer(layer));
    }

    private static Dimension getDimension(JComponent c) {
	if(c == null) return new Dimension(0, 0);
	return c.getPreferredSize();
    }
}
