package lufyt.plot;

import static java.lang.Math.*;
import java.util.Collection;

/**
 * Abstract base-class for calculation of untransformed data-values
 * for major and minor scale-tics. The bounds of the scale are provide
 * by a current Range-reader object.
 */
public abstract class Scale
{
    private Range.Reader range;
    /** Sets the current Range-reader. */
    public void setReader(Range.Reader r) {range = r;}
    /** Apply default-settings to this Scale. */
    public abstract void setTics();
    /** 
     * Calculate major and minor tic-coordinates and store the result
     * in the provided collections. */
    public abstract void getTics(Collection<Double> major,
				 Collection<Double> minor);

    public Scale() {};
    /** Constructor that sets the current Range-reader. */
    public Scale(Range.Reader r) {range = r;}

    static final double ε = 1e-6;

    /** Concretization of Scale with linear spacing between tics. */
    public static class Lin extends Scale {
	protected int mult;
	protected double Δ;
	public Lin() {}
	public Lin(Range.Reader r) {super(r);}
	/** Sets the major tic-distance to step and the minor
	 * tic-distance to step/div. */
	public void setTics(double step, int div) {mult = div; Δ = step/div;}
	/** Set tic-spacing automatically. */
	public void setTics() {
	    double maxStep = 0.4*(super.range.getMax() - super.range.getMin());
	    if(maxStep > 0) {
		double step = exp(log(10)*floor(log(maxStep)/log(10)));
		if(2*step > maxStep)
		    setTics(step, 2);
		else if(5*step > maxStep)
		    setTics(2*step, 2);
		else
		    setTics(5*step, 5);
	    }
	}
	public void getTics(Collection<Double> major, Collection<Double> minor) {
	    minor.clear();
	    major.clear();
	    int begin = (int) ceil(super.range.getMin()/Δ - ε);
	    int n = (int)(super.range.getMax()/Δ + ε - begin);
	    for(int i = begin; i <= begin + n; i++)
		(i % mult == 0? major: minor).add(i*Δ);
	}
    }

    /** Concretization of Scale with logarithmic spacing between tics. */
    public static class Log extends Scale {
	protected double base;
	protected double[] major;
	protected double[] minor;
	public Log() {}
	public Log(Range.Reader r) {super(r);}
	/** Sets logarithmic base for repetitions in the tic-scale and
	 * distributes the tics. The elements of major and minor
	 * should be sorted in ascending order and belong to the
	 * interval [1, base). */
	public void setTics(double base, double[] major, double[] minor) {
	    this.base = base;
	    this.major = (double[]) major.clone();
	    this.minor = (double[]) minor.clone();
	}
	/** Sets the scale according to 10-log with a factor of
	 * 10^step between major tics for positive values of
	 * step. Special denser 10-log scales are generated for step =
	 * 0 and step = -1. */
	public void setTics(int step) {
	    base = 10;
	    if(step < 0) {
		major = new double[] {1, 2, 3, 4, 5, 6, 7, 8, 9};;
		minor = new double[0];
	    }
	    else if(step == 0) {
		major = new double[] {1, 3};
		minor = new double[] {2, 4, 5, 6, 7, 8, 9};;
	    }
	    else if(step == 1) {
		major = new double[] {1};
		minor = new double[] {2, 3, 4, 5, 6, 7, 8, 9};
	    }
	    else {
		major = new double[] {1};
		minor = new double[step - 1];
		for(int i = 1; i < step; i++) {
		    minor[i - 1] = base;
		    base *= 10;
		}
	    }
	}
	/** Calls setTics(1) that generate 10-log scale of common type. */
	public void setTics() {
	    setTics(1);
	}
	public void getTics(Collection<Double> major, Collection<Double> minor) {
	    minor.clear();
	    major.clear();
	    double xMin = super.range.getMin()*(1 - ε);
	    double xMax = super.range.getMax()*(1 + ε);
	    for(double x = pow(base, floor(log(xMin)/log(base)));
		x < xMax; x *= base) {
		for(int i = 0; i < this.major.length; ++i) {
		    double t = x*this.major[i];
		    if(t > xMin && t < xMax)
			major.add(t);
		}
		for(int i = 0; i < this.minor.length; ++i) {
		    double t = x*this.minor[i];
		    if(t > xMin && t < xMax)
			minor.add(t);
		}
	    }
	}
    }

    /** Converts a number to a string for display at major scale-tics. */
    public static String numberString(double x) {
	String str = String.format("%G", x);
	char sep = (new java.text.DecimalFormatSymbols())
	    .getDecimalSeparator();
	char tSep = String.format("%G", 0.1).charAt(1);
	if(sep != tSep) {
	    int i = str.indexOf(tSep);
	    if(i >= 0)
		str = str.substring(0, i) + sep + str.substring(i + 1);
	}
	if(str.indexOf(sep) >= 0) {
	    int e = str.indexOf('E');
	    if(e < 0)
		e = str.length();
	    int i = e;
	    while(i > 1 && str.charAt(i - 1) == '0')
		--i;
	    if(str.charAt(i - 1) == sep)
		--i;
	    if(i < e)
		str = str.substring(0, i) + str.substring(e);
	}
	return str;
    }
}
