import lufyt.plot.*;
import lufyt.gui.Option;
import lufyt.numeric.ODEStepper;
import static lufyt.gui.Option.NumberField;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/**
 * Class for comparisons between numeric integration and analytical
 * calculations for a falling object that is slowed down by viscous
 * and turbulent friction. The class displays a JFrame with a
 * control-panel and a plot-area. The control-panel has text fields
 * for parameter editing and buttons for replotting and selection of
 * plot-types.
 */
public class Freefall extends JFrame implements ActionListener {

    private Plot plot;
    private Parameters param;
    private String paramString;
    private Option.Panel optionPanel;
    private JTextArea statusArea;
    private JLabel[] plotXLabels;
    private JLabel[] plotYLabels;

    public static enum PlotType {
	A_OF_V_PLOT("a(v) plot"),
	V_OF_T_PLOT("v(t) plot"),
	ERROR_PLOT("Error plot");

	public final String id;
	PlotType(String id) {this.id = id;}
	public static PlotType get(String s) {
	    for(PlotType pT: PlotType.values())
		if(s.equals(pT.id)) return pT;
	    return null;
	}
    }

    /** Plain data class for registration of parameters and plot-choice. */
    private class Parameters extends FallParameters {
	String plotType;
	String method;
	double fallT, step0, step1;
    }

    /**
     * Constructor that sets up the graphical interface and displays a
     * plot with default parameter settings.
     */
    public Freefall() {
	super("Fritt fall");

	//Setup plotlabels
	plotXLabels = new JLabel[PlotType.values().length];
	plotYLabels = new JLabel[PlotType.values().length];
	labelSetup(PlotType.A_OF_V_PLOT, "v (m/s)", "a (m/s²)");
	labelSetup(PlotType.V_OF_T_PLOT, "t (s)", "v (m/s)");
	labelSetup(PlotType.ERROR_PLOT, "h (s)", "Δv (m/s)");

	//Setup layout elements
	plot = new Plot();
	plot.setOpaque(true);
	optionPanel = new Option.Panel();
	optionPanel.addActionListener(this);
	optionPanel.setOpaque(true);
	statusArea = new JTextArea();
	statusArea.setLineWrap(true);
	statusArea.setWrapStyleWord(true);
	statusArea.setBorder(new EmptyBorder(2, 4, 2, 4)); 
	plot.setLabelMargin(4);
	plot.setOpaque(true);

	//Setup layout
	setLayout(new BorderLayout());
	Container c = getContentPane();
	c.add(plot, BorderLayout.CENTER);
	c.add(optionPanel, BorderLayout.WEST);
	c.add(statusArea, BorderLayout.SOUTH);
	setDefaultCloseOperation(DISPOSE_ON_CLOSE);
	setPreferredSize(new Dimension(950, 650));
	pack();
	
	//Setup options
	Option.Panel p = optionPanel;
	p.addStretch();
	p.addLabel("Parametrar", new Insets(16, 12, 12, 12));
	p.putOption("gAcc", new NumberField("g (m/s²)", 9.82, 0., true));
	p.putOption("visc", new NumberField("η (Pa\u2009s)", 1.7e-5,
						   0., true));
	p.putOption("density", new NumberField("ρ (kg/m³)", 1.2,
						      0., true));
	p.putOption("mass", new NumberField("m (kg)", 0.058, 0., true));
	p.putOption("radius", new NumberField("r (m)", 0.034, 0., true));
	p.putOption("area", new NumberField("A (m²)", 3.6e-3, 0., true));
	p.putOption("geomC", new NumberField("C (1)", 0.47, 0., true));
	p.putOption("fallT", new NumberField("t₁ (s)", 10., 0., true));
	p.putOption("step0", new NumberField("h₀ (s)", 1e-3, 0., true));
	p.putOption("step1", new NumberField("h₁ (s)", 1e-1, 0., true));
	p.addStretch();
	p.putOption("Method", new Option.Discrete
		    ("Metodval", 8,
		     new String[]{"Analytic", "Euler", "Midpoint"},
		     new String[]{"Analytisk", "Eulersteg", "Mittpunktsmetoden"}
		     ), new Insets(16, 12, 4, 12));
	p.addStretch();
	p.addLabel("Plottning", new Insets(16, 12, 12, 12));
	p.defaultInsets.bottom = 6;
	p.putButton(PlotType.A_OF_V_PLOT.id, new JButton("Plotta a(t)"));
	p.putButton(PlotType.V_OF_T_PLOT.id, new JButton("Plotta v(t)"));
	p.putButton(PlotType.ERROR_PLOT.id, new JButton("Metodavvikelser"));
	p.addStretch();
	p.setBorder(new EmptyBorder(new Insets(12, 0, 12, 0)));

	//Initiate default-plot and show window
	optionPanel.pressButton("a(v) plot");
        setVisible(true);
    }

    public void actionPerformed(ActionEvent e) {
	Option.Panel oP = optionPanel;

	//Transfer plot-settings to local context
	Parameters p = new Parameters();
	p.plotType = e.getActionCommand();
	p.method   = oP.getString("Method");
	p.gAcc     = oP.getNumber("gAcc").doubleValue();
	p.visc     = oP.getNumber("visc").doubleValue();
	p.density  = oP.getNumber("density").doubleValue();
	p.mass     = oP.getNumber("mass").doubleValue();
	p.radius   = oP.getNumber("radius").doubleValue();
	p.area     = oP.getNumber("area").doubleValue();
	p.geomC    = oP.getNumber("geomC").doubleValue();
	p.fallT    = oP.getNumber("fallT").doubleValue();
	p.step0    = oP.getNumber("step0").doubleValue();
	p.step1    = oP.getNumber("step1").doubleValue();

	//Check plot-settings: show error message and return if invalid
	if(p.plotType.equals(PlotType.A_OF_V_PLOT.id) 
	   && !p.method.equals("Analytic")) {
	    oP.errorDialog("a(v) kan endast plottas analytiskt");
	    return;
	}
	if(p.plotType.equals(PlotType.ERROR_PLOT.id)) {
	    if(p.method.equals("Analytic")) {
		oP.errorDialog("Välj en numerisk metod för att plotta"
			       + " avvikelser mot den analytiska lösningen");
		return;
	    }
	    if(p.step0 <= 0 || p.step1 <= 0) {
		oP.errorDialog("Båda steglängder måster vara > 0 för"
			       + " plottning av metodavvikelser");
		return;
	    }
	    if(p.step0 > p.step1) {
		oP.errorDialog ("Steglängderna måste uppfylla h₀ < h₁"
				+ " för plottning av metodavvikelser");
		return;
	    }
	}

	//Transfer plot-settings to this Freefall
	param = p;
	paramString = "";
	String[] opt = oP.getOptionText();
	for(int i = 0; i < opt.length; ++i)
	    paramString += (i > 0? "; ": "") + opt[i];
	
	//Display settings in the status line.
	statusArea.setText(param.plotType + ": " + paramString);

	//Recalculate plot-data, set plot-axis, and display the new plot
	updatePlot();
    }

    /** Updates the plot according to the parameters of this Freefall. */
    private void updatePlot() {
	PlotType pT = PlotType.get(param.plotType);
	plot.setLabelX(plotXLabels[pT.ordinal()]);
	plot.setLabelY(plotYLabels[pT.ordinal()]);
	switch(pT) {
	case A_OF_V_PLOT:
	    plot.setLinX();
	    plot.setLinY();
	    plot.setBoundsX(0., param.gAcc*param.fallT);
	    plot.setBoundsY(0., null);
	    break;
	case V_OF_T_PLOT:
	    plot.setLinX();
	    plot.setLinY();
	    plot.setBoundsX(0., param.fallT);
	    plot.setBoundsY(0., null);
	    break;
	case ERROR_PLOT:
	    plot.setLogX();
	    plot.setLogY();
	    plot.setBoundsX(param.step0, param.step1);
	    plot.setBoundsY(null, null);
	    break;
	default:
	    System.err.println("Invalid plot type-id: " + param.plotType);
	}

	//Get new graphs from a call to getGraphs() and show them
	plot.graphPanel.content.clear();
	Graph[] graphs = getGraphs();
	for(int i = 0; i < graphs.length; ++i)
	    plot.graphPanel.content.add(graphs[i].reader());
	plot.doLayout();
	plot.repaint();
    }

    /** Retrieves new graphs according to the parameters of this Freefall. */
    private Graph[] getGraphs() {
	//This is the place for creation of the graphs that should be plotted.
	//Physical parameters and method-choices are stored in param.

	Graph g[] = null;
	FallODE f = new FallODE(param);
	ODEStepper stepper = null;
/*	if(param.method.equals("Euler")) {
	    //Setup ODEStepper for Euler steps
	    stepper = new SimpleStepper.Euler();
	}
	else if(param.method.equals("Midpoint")) {
	    //Setup ODEStepper for the midpoint method
	    stepper = new SimpleStepper.Midpoint();
	}
*/
	
	//Generate real graphs below
	switch(PlotType.get(param.plotType)) {
	case A_OF_V_PLOT:
	    //Graphic test for exercise 2 
	    g = new Graph[1];
	    g[0] = getAVGraph(f, param.gAcc*param.fallT, 100);
	    g[0].setColor(Color.red);
	    break;
/*	case V_OF_T_PLOT:
	    if(!param.method.equals("Analytic")) {
		//Graphic test for exercise 3
		g = new Graph[2];
		g[0] = getVTGraph(f, stepper, param.step0, param.fallT, 100);
		g[0].setColor(Color.red);
		g[1] = getVTGraph(f, stepper, param.step1, param.fallT, 100);
		g[1].setColor(Color.green);
	    }
	    else {
		//Graphic test for exercise 4
		g = new Graph[1];
		g[0] = getVTGraph(new AnalyticFall(param), param.fallT, 100);
		g[0].setColor(Color.blue);
	    }
	    break;
*/
/*
	case ERROR_PLOT:
	    //Graphic test for exercise 5
	    g = new Graph[1];
	    g[0] = getErrGraph((new AnalyticFall(param)).velocity(param.fallT),
			       f, stepper, param.fallT,
			       param.step0, param.step1, 100);
	    g[0].setColor(Color.blue);
	    break;
*/

	}

	return g;
    }

	private Graph getAVGraph(FallODE f, double vMax, int div) {
		Graph g = new Graph.Line();
		double[] var = new double[1]; 
		for(int i = 0; i <= div; ++i) {
		    var[0] = vMax*i/div;
		    g.add(var[0], f.timeDeriv(var)[0]);
		}
		return g;
	}
	
	private Graph getVTGraph(FallODE f, ODEStepper s, double h,
				     double tMax, int div) {
		Graph g = new Graph.Line();
		double[] var = new double[1];
		int nSteps = (int)(tMax/h + 0.5);
		int gap = (nSteps - 1)/div + 1;
		for(int i = 0; i <= nSteps; ++i) {
		    if(i % gap == 0 || i == nSteps)
			g.add(i*h, var[0]);
		    var = s.step(var, f, h);
		}
		return g;
	}
	
/*	private Graph getVTGraph(AnalyticFall f, double tMax, int div) {
		Graph g = new Graph.Line();
		double[] var = new double[1]; 
		for(int i = 0; i <= div; ++i)
		    g.add(tMax*i/div, f.velocity(tMax*i/div));
		return g;
	}
*/	
	private Graph getErrGraph(double vExact, FallODE f, ODEStepper s,
				      double t, double h0, double h1, int div) {
		Graph g = new Graph.Line();
		double lnH0 = Math.log(h0);
		double lnHF = Math.log(h1) - lnH0;
		for(int i = 0; i <= div; ++i) {
		    int nSteps = (int)(t/Math.exp(lnH0 + lnHF*i/div) + 0.5);
		    double h = t/nSteps;
		    double[] var = new double[1];
		    for(int j = 0; j < nSteps; ++j)
			var = s.step(var, f, h);
		    g.add(h, Math.abs(var[0] - vExact));
		}
		return g;
	}

    /** Called by the constructor and initates plotXLabels and plotYLabels. */
    private void labelSetup(PlotType pType,
			    String xLabel, String yLabel) {
	plotXLabels[pType.ordinal()] = new JLabel(xLabel);
	plotYLabels[pType.ordinal()] = new JLabel(yLabel);
    }
    
    /** Main method that only creates an instance of Freefall */
    public static void main(String[] arg) {
	new Freefall();
    }

}
